/******************************************************************************

  Copyright (c) 2006,2007 Turku PET Centre

  Library:      imgthrs
  Description:  Functions for thresholding and filtering 
                dynamic and parametric PET images.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details:
  http://www.gnu.org/copyleft/lesser.html

  You should have received a copy of the GNU Lesser General Public License
  along with this library/program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 

  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi

  Modification history:
  2006-10-30 Vesa Oikonen
    First created.
  2007-02-01 VO
    imgThresholdTemplate() moved to here (inside the same library).


******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
/*****************************************************************************/
#include "img.h"
#include "imgmax.h"
/*****************************************************************************/
#include "include/imgproc.h"
#include "include/imgthrs.h"
/*****************************************************************************/

/*****************************************************************************/
/** Thresholding the IMG data: pixel time-activity curves (TACs) which have
    AUC less than Threshold*Max_AUC will be set to zero.
\return Returns 0, if ok.
 */
int imgThresholding(
  /** (Dynamic) IMG data */
  IMG *img,
  /** Threshold level, e.g. 0.50 will set to zero all pixels that are less than
      50% of maximum pixel */  
  float threshold_level,
  /** Number of pixels that will fall below the threshold level;
      give NULL pointer, if not needed. */
  int *thr_nr
) {
  int zi, yi, xi, fi;
  int nr=0, ret;
  IMG aucimg;
  float maxauc, thr_limit;
  
  /* Calculate AUC image */
  imgInit(&aucimg);
  ret=imgFrameIntegral(img, 0, img->dimt-1, &aucimg);
  if(ret) return(ret);
  /* Search the max AUC */
  ret=imgMax(&aucimg, &maxauc); if(ret) {imgEmpty(&aucimg); return(ret);}
  /* Calculate the limit value */
  thr_limit=threshold_level*maxauc;
  /* Set to zero the pixels with AUC less than this limit */
  for(zi=0; zi<img->dimz; zi++) for(yi=0; yi<img->dimy; yi++) for(xi=0; xi<img->dimx; xi++)
    if(aucimg.m[zi][yi][xi][0]<thr_limit) {
      for(fi=0; fi<img->dimt; fi++) img->m[zi][yi][xi][fi]=0.0;
      nr++;
    }
  imgEmpty(&aucimg);
  if(thr_nr!=NULL) *thr_nr=nr;
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Filter out pixels that are over limit x higher than their closest 8
    neighbour pixels.
\return Returns the nr of filtered pixels, and negative value in case of
    an error.
 */
int imgOutlierFilter(IMG *img, float limit)
{
  int zi, yi, xi, fi, nr=0;
  float f;
  IMG tmp;

  if(img->status!=IMG_STATUS_OCCUPIED || img->dimt<1) return(-1);
  imgInit(&tmp);
  if(imgAllocate(&tmp, img->dimz, img->dimy, img->dimx, 1)!=0) return(-2);
  /* Process one frame at a time */  
  for(fi=0; fi<img->dimt; fi++) {
    /* Copy one frame to safety */
    for(zi=0; zi<tmp.dimz; zi++)
      for(yi=0; yi<tmp.dimy; yi++)
        for(xi=0; xi<tmp.dimx; xi++)
          tmp.m[zi][yi][xi][0]=img->m[zi][yi][xi][fi];
    /* Go through it */
    for(zi=0; zi<img->dimz; zi++) {
      for(yi=1; yi<img->dimy-1; yi++) for(xi=1; xi<img->dimx-1; xi++) {
        f=tmp.m[zi][yi][xi-1][0]+
          tmp.m[zi][yi][xi+1][0]+
          tmp.m[zi][yi-1][xi][0]+
          tmp.m[zi][yi+1][xi][0]+
          tmp.m[zi][yi+1][xi-1][0]+
          tmp.m[zi][yi+1][xi+1][0]+
          tmp.m[zi][yi-1][xi-1][0]+
          tmp.m[zi][yi-1][xi+1][0];
        f/=8.0;
        if(img->m[zi][yi][xi][fi]>(limit*f)) {img->m[zi][yi][xi][fi]=f; nr++;}
      }
    } /* next plane */
  } /* next frame */
  imgEmpty(&tmp);

  return(nr);
}
/*****************************************************************************/

/*****************************************************************************/
/** Creates a template image based on lower and upper threshold values.
    This function allocates memory for the template.
    If pixel value in original image is >=minValue and <=maxValue, the
    corresponding template pixel is set to 1, otherwise to 0.
    Only the first frame of images is applied.
\return Returns 0 if ok.
 */
int imgThresholdTemplate(
  /** Original image */
  IMG *img,
  /** Lower threshold */
  float minValue,
  /** Upper threshold */
  float maxValue,
  /** Template image */
  IMG *timg
) {
  int fi=0, zi, xi, yi, ret;

  if(img->status!=IMG_STATUS_OCCUPIED) return(1);
  /* Allocate memory for the template data */
  ret=imgAllocate(timg, img->dimz, img->dimy, img->dimx, 1);
  if(ret) return(ret);

  /* set template header information */
  imgCopyhdr(img, timg);
  timg->start[0]=img->start[0]; timg->end[0]=img->end[img->dimt-1];
  timg->mid[0]=(timg->start[0]+timg->end[0])/2.0;

  /* Make template */
  fi=0;
  for(zi=0; zi<img->dimz; zi++)
    for(yi=0; yi<img->dimy; yi++) for(xi=0; xi<img->dimx; xi++) {
      if(img->m[zi][yi][xi][fi]<minValue || img->m[zi][yi][xi][fi]>maxValue) 
        timg->m[zi][yi][xi][fi]=0.0; else timg->m[zi][yi][xi][fi]=1.0;
    }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Threshold IMG by a template.
    Sets pixel values in img to thrValue, if corresponding pixel value in
    template is == 0. Only first plane of template is used.
\return Returns 0, if ok.
 */
int imgThresholdByTemplate(
  /** Image to threshold */
  IMG *img,
  /** Threshold template (1 frame) where 0=cut off */
  IMG *templt,
  /** Value which is written in cut off pixels */
  float thrValue
) {
  int fi, zi, xi, yi;

  if(img->status!=IMG_STATUS_OCCUPIED) return(1);
  if(templt->status!=IMG_STATUS_OCCUPIED) return(2);
  for(zi=0; zi<img->dimz; zi++)
    for(yi=0; yi<img->dimy; yi++) for(xi=0; xi<img->dimx; xi++)
      if(templt->m[zi][yi][xi][0]==0.0) {  
        for(fi=0; fi<img->dimt; fi++) img->m[zi][yi][xi][fi]=thrValue;
      }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/

